from hazelcast.future import ImmediateFuture
from hazelcast.protocol.codec import ringbuffer_add_all_codec, ringbuffer_add_codec, ringbuffer_capacity_codec, \
    ringbuffer_head_sequence_codec, ringbuffer_read_many_codec, ringbuffer_read_one_codec, \
    ringbuffer_remaining_capacity_codec, ringbuffer_size_codec, ringbuffer_tail_sequence_codec
from hazelcast.proxy.base import PartitionSpecificProxy
from hazelcast.util import check_not_negative, check_not_none, check_not_empty, check_true

OVERFLOW_POLICY_OVERWRITE = 0
"""
Configuration property for DEFAULT overflow policy. When an item is tried to be added on full Ringbuffer, oldest item in
the Ringbuffer is overwritten and item is added.
"""
OVERFLOW_POLICY_FAIL = 1
"""
Configuration property for overflow policy. When an item is tried to be added on full Ringbuffer, the call fails and
item is not added.

The reason that FAIL exist is to give the opportunity to obey the ttl. If blocking behavior is required, this can be
implemented using retrying in combination with an exponential backoff.
    >>> sleepMS = 100;
    >>> while true:
    >>>     result = ringbuffer.add(item, -1)
    >>>     if result != -1:
    >>>         break
    >>>     sleep(sleepMS / 1000)
    >>>     sleepMS *= 2
"""
MAX_BATCH_SIZE = 1000
"""
The maximum number of items to be added to RingBuffer or read from RingBuffer at a time.
"""


class Ringbuffer(PartitionSpecificProxy):
    """
    A Ringbuffer is a data-structure where the content is stored in a ring like structure. A Ringbuffer has a capacity
    so it won't grow beyond that capacity and endanger the stability of the system. If that capacity is exceeded, than
    the oldest item in the Ringbuffer is overwritten. The Ringbuffer has 2 always incrementing sequences:
        #. tail_sequence: this is the side where the youngest item is found. So the tail is the side of the Ringbuffer
        where items are added to.
        #. head_sequence: this is the side where the oldest items are found. So the head is the side where items gets
        discarded.
    The items in the Ringbuffer can be found by a sequence that is in between (inclusive) the head and tail sequence.

    A Ringbuffer currently is not a distributed data-structure. So all data is stored in a single partition; comparable
    to the IQueue implementation. But we'll provide an option to partition the data in the near future. A Ringbuffer
    can be used in a similar way as a queue, but one of the key differences is that a queue.take is destructive,
    meaning that only 1 thread is able to take an item. A Ringbuffer.read is not destructive, so you can have multiple
    threads reading the same item multiple times.
    """
    _capacity = None

    def capacity(self):
        """
        Returns the capacity of this Ringbuffer.

        :return: (long), the capacity of Ringbuffer.
        """
        if not self._capacity:
            def cache_capacity(f):
                self._capacity = f.result()
                return f.result()

            return self._encode_invoke(ringbuffer_capacity_codec).continue_with(cache_capacity)
        return ImmediateFuture(self._capacity)

    def size(self):
        """
        Returns number of items in the Ringbuffer.

        :return: (long), the size of Ringbuffer.
        """
        return self._encode_invoke(ringbuffer_size_codec)

    def tail_sequence(self):
        """
        Returns the sequence of the tail. The tail is the side of the Ringbuffer where the items are added to. The
        initial value of the tail is -1.

        :return: (long), the sequence of the tail.
        """
        return self._encode_invoke(ringbuffer_tail_sequence_codec)

    def head_sequence(self):
        """
        Returns the sequence of the head. The head is the side of the Ringbuffer where the oldest items in the
        Ringbuffer are found. If the Ringbuffer is empty, the head will be one more than the tail. The initial value of
        the head is 0 (1 more than tail).

        :return: (long), the sequence of the head.
        """
        return self._encode_invoke(ringbuffer_head_sequence_codec)

    def remaining_capacity(self):
        """
        Returns the remaining capacity of the Ringbuffer.

        :return: (long), the remaining capacity of Ringbuffer.
        """
        return self._encode_invoke(ringbuffer_remaining_capacity_codec)

    def add(self, item, overflow_policy=OVERFLOW_POLICY_OVERWRITE):
        """
        Adds the specified item to the tail of the Ringbuffer. If there is no space in the Ringbuffer, the action is
        determined by overflow policy as :const:`OVERFLOW_POLICY_OVERWRITE` or :const:`OVERFLOW_POLICY_FAIL`.

        :param item: (object), the specified item to be added.
        :param overflow_policy: (int), the OverflowPolicy to be used when there is no space (optional).
        :return: (long), the sequenceId of the added item, or -1 if the add failed.
        """
        return self._encode_invoke(ringbuffer_add_codec, value=self._to_data(item), overflow_policy=overflow_policy)

    def add_all(self, items, overflow_policy=OVERFLOW_POLICY_OVERWRITE):
        """
        Adds all of the item in the specified collection to the tail of the Ringbuffer. An add_all is likely to
        outperform multiple calls to add(object) due to better io utilization and a reduced number of executed
        operations. The items are added in the order of the Iterator of the collection.

        If there is no space in the Ringbuffer, the action is determined by overflow policy as :const:`OVERFLOW_POLICY_OVERWRITE`
        or :const:`OVERFLOW_POLICY_FAIL`.

        :param items: (Collection), the specified collection which contains the items to be added.
        :param overflow_policy: (int), the OverflowPolicy to be used when there is no space (optional).
        :return: (long), the sequenceId of the last written item, or -1 of the last write is failed.
        """
        check_not_empty(items, "items can't be empty")
        if len(items) > MAX_BATCH_SIZE:
            raise AssertionError("Batch size can't be greater than %d" % MAX_BATCH_SIZE)
        for item in items:
            check_not_none(item, "item can't be None")

        item_list = [self._to_data(x) for x in items]
        return self._encode_invoke(ringbuffer_add_all_codec, value_list=item_list, overflow_policy=overflow_policy)

    def read_one(self, sequence):
        """
        Reads one item from the Ringbuffer. If the sequence is one beyond the current tail, this call blocks until an
        item is added. Currently it isn't possible to control how long this call is going to block.

        :param sequence: (long), the sequence of the item to read.
        :return: (object), the read item.
        """
        check_not_negative(sequence, "sequence can't be smaller than 0")
        return self._encode_invoke(ringbuffer_read_one_codec, sequence=sequence)

    def read_many(self, start_sequence, min_count, max_count):
        """
        Reads a batch of items from the Ringbuffer. If the number of available items after the first read item is
        smaller than the max_count, these items are returned. So it could be the number of items read is smaller than
        the max_count. If there are less items available than min_count, then this call blocks. Reading a batch of items
        is likely to perform better because less overhead is involved.

        :param start_sequence: (long),  the start_sequence of the first item to read.
        :param min_count: (int), the minimum number of items to read.
        :param max_count: (int), the maximum number of items to read.
        :return: (Sequence), the list of read items.
        """
        check_not_negative(start_sequence, "sequence can't be smaller than 0")
        check_true(max_count >= min_count, "max count should be greater or equal to min count")
        check_true(min_count <= self.capacity().result(), "min count should be smaller or equal to capacity")
        check_true(max_count < MAX_BATCH_SIZE, "max count can't be greater than %d" % MAX_BATCH_SIZE)

        return self._encode_invoke(ringbuffer_read_many_codec, response_handler=self._read_many_response_handler,
                                   start_sequence=start_sequence, min_count=min_count,
                                   max_count=max_count, filter=None)

    @staticmethod
    def _read_many_response_handler(future, codec, to_object):
        return codec.decode_response(future.result(), to_object)['items']



from hazelcast.proxy.base import Proxy


class ReliableTopic(Proxy):
    def add_listener(self, on_message=None):
        raise NotImplementedError

    def publish(self, message):
        raise NotImplementedError

    def remove_listener(self, registration_id):
        raise NotImplementedError



import asyncore
import errno
import logging
import select
import socket
import sys
import threading
import time

from collections import deque
from functools import total_ordering
from hazelcast.config import PROTOCOL
from hazelcast.connection import Connection, BUFFER_SIZE
from hazelcast.exception import HazelcastError
from hazelcast.future import Future
from hazelcast.six.moves import queue

try:
    import ssl
except ImportError:
    ssl = None


class AsyncoreReactor(object):
    _thread = None
    _is_live = False
    logger = logging.getLogger("HazelcastClient.AsyncoreReactor")

    def __init__(self, logger_extras=None):
        self._logger_extras = logger_extras
        self._timers = queue.PriorityQueue()
        self._map = {}

    def start(self):
        self._is_live = True
        self._thread = threading.Thread(target=self._loop, name="hazelcast-reactor")
        self._thread.daemon = True
        self._thread.start()

    def _loop(self):
        self.logger.debug("Starting Reactor Thread", extra=self._logger_extras)
        Future._threading_locals.is_reactor_thread = True
        while self._is_live:
            try:
                asyncore.loop(count=1, timeout=0.01, map=self._map)
                self._check_timers()
            except select.error as err:
                # TODO: parse error type to catch only error "9"
                self.logger.warning("Connection closed by server", extra=self._logger_extras)
                pass
            except:
                self.logger.exception("Error in Reactor Thread", extra=self._logger_extras)
                # TODO: shutdown client
                return
        self.logger.debug("Reactor Thread exited. %s" % self._timers.qsize(), extra=self._logger_extras)
        self._cleanup_all_timers()

    def _check_timers(self):
        now = time.time()
        while not self._timers.empty():
            try:
                _, timer = self._timers.queue[0]
            except IndexError:
                return

            if timer.check_timer(now):
                try:
                    self._timers.get_nowait()
                except queue.Empty:
                    pass
            else:
                return

    def add_timer_absolute(self, timeout, callback):
        timer = Timer(timeout, callback, self._cleanup_timer)
        self._timers.put_nowait((timer.end, timer))
        return timer

    def add_timer(self, delay, callback):
        return self.add_timer_absolute(delay + time.time(), callback)

    def shutdown(self):
        if not self._is_live:
            return
        self._is_live = False
        for connection in list(self._map.values()):
            try:
                connection.close(HazelcastError("Client is shutting down"))
            except OSError as connection:
                if connection.args[0] == socket.EBADF:
                    pass
                else:
                    raise
        self._map.clear()
        self._thread.join()

    def new_connection(self, address, connect_timeout, socket_options, connection_closed_callback, message_callback,
                       network_config):
        return AsyncoreConnection(self._map, address, connect_timeout, socket_options,
                                  connection_closed_callback, message_callback, network_config, self._logger_extras)

    def _cleanup_timer(self, timer):
        try:
            self.logger.debug("Cancel timer %s" % timer, extra=self._logger_extras)
            self._timers.queue.remove((timer.end, timer))
        except ValueError:
            pass

    def _cleanup_all_timers(self):
        while not self._timers.empty():
            try:
                _, timer = self._timers.get_nowait()
                timer.timer_ended_cb()
            except queue.Empty:
                return


class AsyncoreConnection(Connection, asyncore.dispatcher):
    sent_protocol_bytes = False
    read_buffer_size = BUFFER_SIZE

    def __init__(self, map, address, connect_timeout, socket_options, connection_closed_callback,
                 message_callback, network_config, logger_extras=None):
        asyncore.dispatcher.__init__(self, map=map)
        Connection.__init__(self, address, connection_closed_callback, message_callback, logger_extras)

        self._write_lock = threading.Lock()
        self._write_queue = deque()
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(connect_timeout)

        # set tcp no delay
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        # set socket buffer
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, BUFFER_SIZE)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, BUFFER_SIZE)

        for socket_option in socket_options:
            if socket_option.option is socket.SO_RCVBUF:
                self.read_buffer_size = socket_option.value

            self.socket.setsockopt(socket_option.level, socket_option.option, socket_option.value)

        self.connect(self._address)

        ssl_config = network_config.ssl_config
        if ssl and ssl_config.enabled:
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

            protocol = ssl_config.protocol

            # Use only the configured protocol
            try:
                if protocol != PROTOCOL.SSLv2:
                    ssl_context.options |= ssl.OP_NO_SSLv2
                if protocol != PROTOCOL.SSLv3 and protocol != PROTOCOL.SSL:
                    ssl_context.options |= ssl.OP_NO_SSLv3
                if protocol != PROTOCOL.TLSv1:
                    ssl_context.options |= ssl.OP_NO_TLSv1
                if protocol != PROTOCOL.TLSv1_1:
                    ssl_context.options |= ssl.OP_NO_TLSv1_1
                if protocol != PROTOCOL.TLSv1_2 and protocol != PROTOCOL.TLS:
                    ssl_context.options |= ssl.OP_NO_TLSv1_2
                if protocol != PROTOCOL.TLSv1_3:
                    ssl_context.options |= ssl.OP_NO_TLSv1_3
            except AttributeError:
                pass

            ssl_context.verify_mode = ssl.CERT_REQUIRED

            if ssl_config.cafile:
                ssl_context.load_verify_locations(ssl_config.cafile)
            else:
                ssl_context.load_default_certs()

            if ssl_config.certfile:
                ssl_context.load_cert_chain(ssl_config.certfile, ssl_config.keyfile, ssl_config.password)

            if ssl_config.ciphers:
                ssl_context.set_ciphers(ssl_config.ciphers)

            self.socket = ssl_context.wrap_socket(self.socket)

        # the socket should be non-blocking from now on
        self.socket.settimeout(0)

        self._write_queue.append(b"CB2")

    def handle_connect(self):
        self.start_time_in_seconds = time.time()
        self.logger.debug("Connected to %s", self._address, extra=self._logger_extras)

    def handle_read(self):
        self._read_buffer.extend(self.recv(self.read_buffer_size))
        self.last_read_in_seconds = time.time()
        self.receive_message()

    def handle_write(self):
        with self._write_lock:
            try:
                data = self._write_queue.popleft()
            except IndexError:
                return
            sent = self.send(data)
            self.last_write_in_seconds = time.time()
            self.sent_protocol_bytes = True
            if sent < len(data):
                self._write_queue.appendleft(data[sent:])

    def handle_close(self):
        self.logger.warning("Connection closed by server", extra=self._logger_extras)
        self.close(IOError("Connection closed by server"))

    def handle_error(self):
        error = sys.exc_info()[1]
        if sys.exc_info()[0] is socket.error:
            if error.errno != errno.EAGAIN and error.errno != errno.EDEADLK:
                self.logger.exception("Received error", extra=self._logger_extras)
                self.close(IOError(error))
        else:
            self.logger.warning("Received unexpected error: " + str(error), extra=self._logger_extras)

    def readable(self):
        return not self._closed and self.sent_protocol_bytes

    def write(self, data):
        # if write queue is empty, send the data right away, otherwise add to queue
        if len(self._write_queue) == 0 and self._write_lock.acquire(False):
            try:
                sent = self.send(data)
                self.last_write_in_seconds = time.time()
                if sent < len(data):
                    self.logger.info("Adding to queue", extra=self._logger_extras)
                    self._write_queue.appendleft(data[sent:])
            finally:
                self._write_lock.release()
        else:
            self._write_queue.append(data)

    def writable(self):
        return len(self._write_queue) > 0

    def close(self, cause):
        if not self._closed:
            self._closed = True
            asyncore.dispatcher.close(self)
            self._connection_closed_callback(self, cause)


@total_ordering
class Timer(object):
    canceled = False

    def __init__(self, end, timer_ended_cb, timer_canceled_cb):
        self.end = end
        self.timer_ended_cb = timer_ended_cb
        self.timer_canceled_cb = timer_canceled_cb

    def __eq__(self, other):
        return self.end == other.end

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return self.end < other.end

    def cancel(self):
        self.canceled = True
        self.timer_canceled_cb(self)

    def check_timer(self, now):
        if self.canceled:
            return True

        if now > self.end:
            self.timer_ended_cb()
            return True



"""
Hazelcast Client Configuration module contains configuration classes and various constants required to create a ClientConfig.

"""
import logging
import os

from hazelcast.serialization.api import StreamSerializer
from hazelcast.util import validate_type, validate_serializer, enum, TimeUnit

DEFAULT_GROUP_NAME = "dev"
"""
Default group name of the connected Hazelcast cluster
"""

DEFAULT_GROUP_PASSWORD = "dev-pass"
"""
Default password of connected Hazelcast cluster
"""

INTEGER_TYPE = enum(VAR=0, BYTE=1, SHORT=2, INT=3, LONG=4, BIG_INT=5)
"""
Integer type options that can be used by serialization service.

* VAR : variable size integer (this option can be problematic on static type clients like java or .NET)
* BYTE: Python int will be interpreted as a single byte int
* SHORT: Python int will be interpreted as a double byte int
* INT: Python int will be interpreted as a four byte int
* LONG: Python int will be interpreted as an eight byte int
* BIG_INT: Python int will be interpreted as Java BigInteger. This option can handle python long values with "bit_length > 64"

"""

EVICTION_POLICY = enum(NONE=0, LRU=1, LFU=2, RANDOM=3)
"""
Near Cache eviction policy options

* NONE : No evcition
* LRU : Least Recently Used items will be evicted
* LFU : Least frequently Used items will be evicted
* RANDOM : Items will be evicted randomly

"""

IN_MEMORY_FORMAT = enum(BINARY=0, OBJECT=1)
"""
Near Cache in memory format of the values.

* BINARY : Binary format, hazelcast serializated bytearray format
* OBJECT : The actual objects used

"""

PROTOCOL = enum(SSLv2=0, SSLv3=1, SSL=2, TLSv1=3, TLSv1_1=4, TLSv1_2=5, TLSv1_3=6, TLS=7)
"""
SSL protocol options.

* SSLv2     : SSL 2.O Protocol. RFC 6176 prohibits SSL 2.0. Please use TLSv1+
* SSLv3     : SSL 3.0 Protocol. RFC 7568 prohibits SSL 3.0. Please use TLSv1+
* SSL       : Alias for SSL 3.0
* TLSv1     : TLS 1.0 Protocol described in RFC 2246
* TLSv1_1   : TLS 1.1 Protocol described in RFC 4346
* TLSv1_2   : TLS 1.2 Protocol described in RFC 5246
* TLSv1_3   : TLS 1.3 Protocol described in RFC 8446
* TLS       : Alias for TLS 1.2
* TLSv1+ requires at least Python 2.7.9 or Python 3.4 build with OpenSSL 1.0.1+ 
* TLSv1_3 requires at least Python 2.7.15 or Python 3.7 build with OpenSSL 1.1.1+
"""

DEFAULT_MAX_ENTRY_COUNT = 10000
DEFAULT_SAMPLING_COUNT = 8
DEFAULT_SAMPLING_POOL_SIZE = 16

MAXIMUM_PREFETCH_COUNT = 100000


class ClientConfig(object):
    """
    The root configuration for hazelcast python client.

        >>> client_config = ClientConfig()
        >>> client = HazelcastClient(client_config)
    """

    def __init__(self):
        self._properties = {}
        """Config properties"""

        self.group_config = GroupConfig()
        """The group configuration"""

        self.network_config = ClientNetworkConfig()
        """The network configuration for addresses to connect, smart-routing, socket-options..."""

        self.load_balancer = None
        """Custom load balancer used to distribute the operations to multiple Endpoints."""

        self.membership_listeners = []
        """Membership listeners, an array of tuple (member_added, member_removed, fire_for_existing)"""

        self.lifecycle_listeners = []
        """ Lifecycle Listeners, an array of Functions of f(state)"""

        self.near_cache_configs = {}  # map_name:NearCacheConfig
        """Near Cache configuration which maps "map-name : NearCacheConfig"""

        self.flake_id_generator_configs = {}
        """Flake ID generator configuration which maps "config-name" : FlakeIdGeneratorConfig """

        self.serialization_config = SerializationConfig()
        """Hazelcast serialization configuration"""

        self.logger_config = LoggerConfig()
        """Logger configuration."""

        self.client_name = ""
        """Name of the client"""

    def add_membership_listener(self, member_added=None, member_removed=None, fire_for_existing=False):
        """
        Helper method for adding membership listeners

        :param member_added: (Function), Function to be called when a member is added, in the form of f(member)
        (optional).
        :param member_removed: (Function), Function to be called when a member is removed, in the form of f(member)
        (optional).
        :param fire_for_existing: if True, already existing members will fire member_added event (optional).
        :return: `self` for cascading configuration
        """
        self.membership_listeners.append((member_added, member_removed, fire_for_existing))
        return self

    def add_lifecycle_listener(self, lifecycle_state_changed=None):
        """
        Helper method for adding lifecycle listeners.

        :param lifecycle_state_changed: (Function), Function to be called when lifecycle state is changed (optional).
        In the form of f(state).
        :return: `self` for cascading configuration
        """
        if lifecycle_state_changed:
            self.lifecycle_listeners.append(lifecycle_state_changed)
        return self

    def add_near_cache_config(self, near_cache_config):
        """
        Helper method to add a new NearCacheConfig.

        :param near_cache_config: (NearCacheConfig), the near_cache config to add.
        :return: `self` for cascading configuration.
        """
        self.near_cache_configs[near_cache_config.name] = near_cache_config
        return self

    def add_flake_id_generator_config(self, flake_id_generator_config):
        """
        Helper method to add a new FlakeIdGeneratorConfig.

        :param flake_id_generator_config: (FlakeIdGeneratorConfig), the configuration to add
        :return: `self` for cascading configuration
        """
        self.flake_id_generator_configs[flake_id_generator_config.name] = flake_id_generator_config
        return self

    def get_property_or_default(self, key, default):
        """
        Client property accessor with fallback to default value.

        :param key: (Object), property key to access.
        :param default: (Object), the default value for fallback.
        :return: (Object), property value if it exist or the default value otherwise.
        """
        try:
            return self._properties[key]
        except KeyError:
            return default

    def get_properties(self):
        """
        Gets the configuration properties.

        :return: (dict), Client configuration properties.
        """
        return self._properties

    def set_property(self, key, value):
        """
        Sets the value of a named property.

        :param key: Property name
        :param value: Value of the property
        :return: `self` for cascading configuration.
        """
        self._properties[key] = value
        return self


class GroupConfig(object):
    """
    The Group Configuration is the container class for name and password of the cluster.
    """

    def __init__(self):
        self.name = DEFAULT_GROUP_NAME
        """The group name of the cluster"""
        self.password = DEFAULT_GROUP_PASSWORD
        """The password of the cluster"""


class ClientNetworkConfig(object):
    """
    Network related configuration parameters.
    """

    def __init__(self):
        self.addresses = []
        """The candidate address list that client will use to establish initial connection"""
        """Example usage: addresses.append("127.0.0.1:5701") """
        self.connection_attempt_limit = 2
        """
        While client is trying to connect initially to one of the members in the addressList, all might be not
        available. Instead of giving up, throwing Error and stopping client, it will attempt to retry as much as defined
        by this parameter.
        """
        self.connection_attempt_period = 3
        """Period for the next attempt to find a member to connect"""
        self.connection_timeout = 5.0
        """
        Socket connection timeout is a float, giving in seconds, or None.
        Setting a timeout of None disables the timeout feature and is equivalent to block the socket until it connects.
        Setting a timeout of zero is the same as disables blocking on connect.
        """
        self.socket_options = []
        """
        Array of Unix socket options.

        Example usage:

            >>> import socket
            >>> client_network_config.socket_options.append(SocketOption(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1))
            >>> client_network_config.socket_options.append(SocketOption(socket.SOL_SOCKET, socket.SO_SNDBUF, 32768))
            >>> client_network_config.socket_options.append(SocketOption(socket.SOL_SOCKET, socket.SO_RCVBUF, 32768))

        Please see the Unix manual for level and option. Level and option constant are in python std lib socket module
        """
        self.redo_operation = False
        """
        If true, client will redo the operations that were executing on the server and client lost the connection.
        This can be because of network, or simply because the member died. However it is not clear whether the
        application is performed or not. For idempotent operations this is harmless, but for non idempotent ones
        retrying can cause to undesirable effects. Note that the redo can perform on any member.
        """
        self.smart_routing = True
        """
        If true, client will route the key based operations to owner of the key at the best effort. Note that it uses a
        cached value of partition count and doesn't guarantee that the operation will always be executed on the owner.
        The cached table is updated every 10 seconds.
        """
        self.ssl_config = SSLConfig()
        """SSL configurations for the client."""
        self.cloud_config = ClientCloudConfig()
        """Hazelcast Cloud configuration to let the client connect the cluster via Hazelcast.cloud"""


class SocketOption(object):
    """
    Advanced configuration for fine-tune the TCP options.
    A Socket option represent the unix socket option, that will be passed to python socket.setoption(level,`option, value)`
    See the Unix manual for level and option.
    """

    def __init__(self, level, option, value):
        self.level = level
        """Option level. See the Unix manual for detail."""
        self.option = option
        """The actual socket option. The actual socket option."""
        self.value = value
        """Socket option value. The value argument can either be an integer or a string"""


class SerializationConfig(object):
    """
    Hazelcast Serialization Service configuration options can be set from this class.
    """

    def __init__(self):
        self.portable_version = 0
        """
        Portable version will be used to differentiate two versions of the same class that have changes on the class,
        like adding/removing a field or changing a type of a field.
        """
        self.data_serializable_factories = {}
        """
        Dictionary of factory-id and corresponding IdentifiedDataserializable factories. A Factory is a simple
        dictionary with entries of class-id : class-constructor-function pairs.

        Example:

            >>> my_factory = {MyPersonClass.CLASS_ID : MyPersonClass, MyAddressClass.CLASS_ID : MyAddressClass}
            >>> serialization_config.data_serializable_factories[FACTORY_ID] = my_factory

        """
        self.portable_factories = {}
        """
        Dictionary of factory-id and corresponding portable factories. A Factory is a simple dictionary with entries of
        class-id : class-constructor-function pairs.

        Example:

            >>> portable_factory = {PortableClass_0.CLASS_ID : PortableClass_0, PortableClass_1.CLASS_ID : PortableClass_1}
            >>> serialization_config.portable_factories[FACTORY_ID] = portable_factory
        """
        self.class_definitions = set()
        """
        Set of all Portable class definitions.
        """
        self.check_class_def_errors = True
        """Configured Portable Class definitions should be validated for errors or not."""
        self.is_big_endian = True
        """Hazelcast Serialization is big endian or not."""
        self.default_integer_type = INTEGER_TYPE.INT
        """
        Python has variable length int/long type. In order to match this with static fixed length Java server, this option
        defines the length of the int/long.
        One of the values of :const:`INTEGER_TYPE` can be assigned. Please see :const:`INTEGER_TYPE` documentation for details of the options.
        """
        self._global_serializer = None
        self._custom_serializers = {}

    def add_data_serializable_factory(self, factory_id, factory):
        """
        Helper method for adding IdentifiedDataSerializable factory.
        example:
            >>> my_factory = {MyPersonClass.CLASS_ID : MyPersonClass, MyAddressClass.CLASS_ID : MyAddressClass}
            >>> serialization_config.add_data_serializable_factory(factory_id, my_factory)

        :param factory_id: (int), factory-id to register.
        :param factory: (Dictionary), the factory dictionary of class-id:class-constructor-function pairs.
        """
        self.data_serializable_factories[factory_id] = factory

    def add_portable_factory(self, factory_id, factory):
        """
        Helper method for adding Portable factory.
        example:
            >>> portable_factory = {PortableClass_0.CLASS_ID : PortableClass_0, PortableClass_1.CLASS_ID : PortableClass_1}
            >>> serialization_config.portable_factories[FACTORY_ID] = portable_factory

        :param factory_id: (int), factory-id to register.
        :param factory: (Dictionary), the factory dictionary of class-id:class-constructor-function pairs.
        """
        self.portable_factories[factory_id] = factory

    def set_custom_serializer(self, _type, serializer):
        """
        Assign a serializer for the type.

        :param _type: (Type), the target type of the serializer
        :param serializer: (Serializer), Custom Serializer constructor function
        """
        validate_type(_type)
        validate_serializer(serializer, StreamSerializer)
        self._custom_serializers[_type] = serializer

    @property
    def custom_serializers(self):
        """
        All custom serializers.

        :return: (Dictionary), dictionary of type-custom serializer pairs.
        """
        return self._custom_serializers

    @property
    def global_serializer(self):
        """
        The Global serializer property for serialization service. The assigned value should be a class constructor
        function. It handles every object if no other serializer found.

        Global serializers should extend `hazelcast.serializer.api.StreamSerializer`
        """
        return self._global_serializer

    @global_serializer.setter
    def global_serializer(self, global_serializer):
        validate_serializer(global_serializer, StreamSerializer)
        self._global_serializer = global_serializer


class NearCacheConfig(object):
    """
    Map Near cache configuration for a specific map by name.
    """

    def __init__(self, name="default"):
        self._name = name
        self.invalidate_on_change = True
        """Should a value is invalidated and removed in case of any map data updating operations such as replace, remove etc."""
        self._in_memory_format = IN_MEMORY_FORMAT.BINARY
        self._time_to_live_seconds = None
        self._max_idle_seconds = None
        self._eviction_policy = EVICTION_POLICY.NONE
        self._eviction_max_size = DEFAULT_MAX_ENTRY_COUNT
        self._eviction_sampling_count = DEFAULT_SAMPLING_COUNT
        self._eviction_sampling_pool_size = DEFAULT_SAMPLING_POOL_SIZE

    @property
    def name(self):
        """Name of the map that this near cache belong. Cannot be None."""
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            raise ValueError("Name of the map cannot be None")
        self._name = name

    @property
    def in_memory_format(self):
        """Internal representation of the stored data in near cache."""
        return self._in_memory_format

    @in_memory_format.setter
    def in_memory_format(self, in_memory_format=IN_MEMORY_FORMAT.BINARY):
        if in_memory_format not in IN_MEMORY_FORMAT.reverse:
            raise ValueError("Invalid in-memory-format :{}".format(in_memory_format))
        self._in_memory_format = in_memory_format

    @property
    def time_to_live_seconds(self):
        """The maximum number of seconds for each entry to stay in the near cache."""
        return self._time_to_live_seconds

    @time_to_live_seconds.setter
    def time_to_live_seconds(self, time_to_live_seconds):
        if time_to_live_seconds < 0:
            raise ValueError("'time_to_live_seconds' cannot be less than 0")
        self._time_to_live_seconds = time_to_live_seconds

    @property
    def max_idle_seconds(self):
        """Maximum number of seconds each entry can stay in the near cache as untouched (not-read)."""
        return self._max_idle_seconds

    @max_idle_seconds.setter
    def max_idle_seconds(self, max_idle_seconds):
        if max_idle_seconds < 0:
            raise ValueError("'max_idle_seconds' cannot be less than 0")
        self._max_idle_seconds = max_idle_seconds

    @property
    def eviction_policy(self):
        """The eviction policy for the near cache"""
        return self._eviction_policy

    @eviction_policy.setter
    def eviction_policy(self, eviction_policy):
        if eviction_policy not in EVICTION_POLICY.reverse:
            raise ValueError("Invalid eviction_policy :{}".format(eviction_policy))
        self._eviction_policy = eviction_policy

    @property
    def eviction_max_size(self):
        """The limit for number of entries until the eviction start."""
        return self._eviction_max_size

    @eviction_max_size.setter
    def eviction_max_size(self, eviction_max_size):
        if eviction_max_size < 1:
            raise ValueError("'Eviction-max-size' cannot be less than 1")
        self._eviction_max_size = eviction_max_size

    @property
    def eviction_sampling_count(self):
        """The entry count of the samples for the internal eviction sampling algorithm taking samples in each
        operation."""
        return self._eviction_sampling_count

    @eviction_sampling_count.setter
    def eviction_sampling_count(self, eviction_sampling_count):
        if eviction_sampling_count < 1:
            raise ValueError("'eviction_sampling_count' cannot be less than 1")
        self._eviction_sampling_count = eviction_sampling_count

    @property
    def eviction_sampling_pool_size(self):
        """The size of the internal eviction sampling algorithm has a pool of best candidates for eviction."""
        return self._eviction_sampling_pool_size

    @eviction_sampling_pool_size.setter
    def eviction_sampling_pool_size(self, eviction_sampling_pool_size):
        if eviction_sampling_pool_size < 1:
            raise ValueError("'eviction_sampling_pool_size' cannot be less than 1")
        self._eviction_sampling_pool_size = eviction_sampling_pool_size


class SSLConfig(object):
    """
    SSL configuration.
    """

    def __init__(self):
        self.enabled = False
        """Enables/disables SSL."""

        self.cafile = None
        """
        Absolute path of concatenated CA certificates used to validate server's certificates in PEM format.
        When SSL is enabled and cafile is not set, a set of default CA certificates from default locations
        will be used.
        """

        self.certfile = None
        """Absolute path of the client certificate in PEM format."""

        self.keyfile = None
        """
        Absolute path of the private key file for the client certificate in the PEM format.
        If this parameter is None, private key will be taken from certfile.
        """

        self.password = None
        """
        Password for decrypting the keyfile if it is encrypted.
        The password may be a function to call to get the password.
        It will be called with no arguments, and it should return a string, bytes, or bytearray.
        If the return value is a string it will be encoded as UTF-8 before using it to decrypt the key.
        Alternatively a string, bytes, or bytearray value may be supplied directly as the password.
        """

        self.protocol = PROTOCOL.TLS
        """Protocol version used in SSL communication. Default value is TLSv1.2"""

        self.ciphers = None
        """
        String in the OpenSSL cipher list format to set the available ciphers for sockets.
        More than one cipher can be set by separating them with a colon.
        """


class FlakeIdGeneratorConfig(object):
    """
    FlakeIdGeneratorConfig contains the configuration for the client regarding
    :class:`~hazelcast.proxy.flake_id_generator.FlakeIdGenerator`
    """

    def __init__(self, name="default"):
        self._name = name
        self._prefetch_count = 100
        self._prefetch_validity_in_millis = 600000

    @property
    def name(self):
        """
        Name of the flake ID generator configuration.

        :getter: Returns the configuration name. This can be actual object name or pattern.
        :setter: Sets the name or name pattern for this config. Must not be modified after this
            instance is added to configuration.
        :type: str
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def prefetch_count(self):
        """
        Prefetch value count.

        :getter: Returns the prefetch count. Prefetch count is in the range 1..100,000.
        :setter: Sets how many IDs are pre-fetched on the background when a new flake ID is requested
            from members. Default is 100.
            Prefetch count should be in the range 1..100,000.
        :type: int
        """
        return self._prefetch_count

    @prefetch_count.setter
    def prefetch_count(self, prefetch_count):
        if not (0 < prefetch_count <= MAXIMUM_PREFETCH_COUNT):
            raise ValueError("Prefetch count must be 1..{}, not {}".format(MAXIMUM_PREFETCH_COUNT, prefetch_count))
        self._prefetch_count = prefetch_count

    @property
    def prefetch_validity_in_millis(self):
        """
        Prefetch validity in milliseconds.

        :getter: Returns the prefetch validity in milliseconds.
        :setter: Sets for how long the pre-fetched IDs can be used.
            If this time elapses, a new batch of IDs will be fetched.
            Time unit is milliseconds, default is 600,000 (10 minutes).
            The IDs contain timestamp component, which ensures rough global ordering of IDs.
            If an ID is assigned to an object that was created much later, it will be much out of order.
            If you don't care about ordering, set this value to 0.
        :type: int
        """
        return self._prefetch_validity_in_millis

    @prefetch_validity_in_millis.setter
    def prefetch_validity_in_millis(self, prefetch_validity_in_millis):
        self._prefetch_validity_in_millis = prefetch_validity_in_millis


class ClientCloudConfig(object):
    """
    Hazelcast Cloud configuration to let the client connect the cluster via Hazelcast.cloud
    """

    def __init__(self):
        self.enabled = False
        """Enables/disables cloud config."""
        self.discovery_token = ""
        """Hazelcast Cloud Discovery token of your cluster."""


class LoggerConfig(object):
    """
    Custom configuration for logging or a logging level for the default
    Hazelcast client logger can be set using this class.
    """
    def __init__(self):
        self.config_file = None
        """
        If the configuration file is set, given configuration file
        will be used instead of the default logger configuration
        with the given log level. This should be the absolute 
        path of a JSON file that follows the 
        ``Configuration dictionary schema`` described in the logging 
        module of the standard library.
        """
        self.level = logging.INFO
        """
        Sets the logging level for the default logging
        configuration. To turn off the logging, level
        can be set to a high integer value. If custom
        logging levels are not used, a value greater 
        than 50 is enough to turn off the default
        logger. 
        """


class ClientProperty(object):
    """
    Client property holds the name, default value and time unit of Hazelcast client properties.
    Client properties can be set by

    * Programmatic Configuration
    * Environment variables
    """

    def __init__(self, name, default_value=None, time_unit=None):
        self.name = name
        self.default_value = default_value
        self.time_unit = time_unit


class ClientProperties(object):
    HEARTBEAT_INTERVAL = ClientProperty("hazelcast.client.heartbeat.interval", 5000, TimeUnit.MILLISECOND)
    """
    Time interval between the heartbeats sent by the client to the nodes.
    """

    HEARTBEAT_TIMEOUT = ClientProperty("hazelcast.client.heartbeat.timeout", 60000, TimeUnit.MILLISECOND)
    """
    Client sends heartbeat messages to the members and this is the timeout for this sending operations.
    If there is not any message passing between the client and member within the given time via this property
    in milliseconds, the connection will be closed.
    """

    INVOCATION_TIMEOUT_SECONDS = ClientProperty("hazelcast.client.invocation.timeout.seconds", 120, TimeUnit.SECOND)
    """
    When an invocation gets an exception because
    
    * Member throws an exception.
    * Connection between the client and member is closed.
    * Client's heartbeat requests are timed out.
    
    Time passed since invocation started is compared with this property.
    If the time is already passed, then the exception is delegated to the user. If not, the invocation is retried.
    Note that, if invocation gets no exception and it is a long running one, then it will not get any exception,
    no matter how small this timeout is set.
    """

    INVOCATION_RETRY_PAUSE_MILLIS = ClientProperty("hazelcast.client.invocation.retry.pause.millis", 1000,
                                                   TimeUnit.MILLISECOND)
    """
    Pause time between each retry cycle of an invocation in milliseconds.
    """

    HAZELCAST_CLOUD_DISCOVERY_TOKEN = ClientProperty("hazelcast.client.cloud.discovery.token", "")
    """
    Token to use when discovering cluster via Hazelcast.cloud.
    """

    STATISTICS_ENABLED = ClientProperty("hazelcast.client.statistics.enabled", False)
    """
    Used to enable the client statistics collection.
    """

    STATISTICS_PERIOD_SECONDS = ClientProperty("hazelcast.client.statistics.period.seconds", 3, TimeUnit.SECOND)
    """
    Period in seconds to collect statistics.
    """

    SERIALIZATION_INPUT_RETURNS_BYTEARRAY = ClientProperty("hazelcast.serialization.input.returns.bytearray", False)
    """
    Input#read_byte_array returns a List if property is False, otherwise it will return a byte-array.
    Changing this to True, gives a considerable performance benefit.
    """

    def __init__(self, properties):
        self._properties = properties

    def get(self, property):
        """
        Gets the value of the given property. First checks client config properties, then environment variables
        and lastly fall backs to the default value of the property.

        :param property: (:class:`~hazelcast.config.ClientProperty`), Property to get value from
        :return: Value of the given property
        """
        value = self._properties.get(property.name, None)
        if value is not None:
            return value

        value = os.getenv(property.name, None)
        if value is not None:
            return value

        return property.default_value

    def get_bool(self, property):
        """
        Gets the value of the given property as boolean.

        :param property: (:class:`~hazelcast.config.ClientProperty`), Property to get value from
        :return: (bool), Value of the given property
        """
        value = self.get(property)
        if isinstance(value, bool):
            return value
        return value.lower() == "true"

    def get_seconds(self, property):
        """
        Gets the value of the given property in seconds. If the value of the given property is not a number,
        throws TypeError.

        :param property: (:class:`~hazelcast.config.ClientProperty`), Property to get seconds from
        :return: (float), Value of the given property in seconds
        """
        return TimeUnit.to_seconds(self.get(property), property.time_unit)

    def get_seconds_positive_or_default(self, property):
        """
        Gets the value of the given property in seconds. If the value of the given property is not a number,
        throws TypeError. If the value of the given property in seconds is not positive, tries to
        return the default value in seconds.

        :param property: (:class:`~hazelcast.config.ClientProperty`), Property to get seconds from
        :return: (float), Value of the given property in seconds if it is positive.
            Else, value of the default value of given property in seconds.
        """
        seconds = self.get_seconds(property)
        return seconds if seconds > 0 else TimeUnit.to_seconds(property.default_value, property.time_unit)



from hazelcast.protocol.codec import \
    topic_add_message_listener_codec, \
    topic_publish_codec, \
    topic_remove_message_listener_codec
from hazelcast.proxy.base import PartitionSpecificProxy, TopicMessage


class Topic(PartitionSpecificProxy):
    """
    Hazelcast provides distribution mechanism for publishing messages that are delivered to multiple subscribers, which
    is also known as a publish/subscribe (pub/sub) messaging model. Publish and subscriptions are cluster-wide. When a
    member subscribes for a topic, it is actually registering for messages published by any member in the cluster,
    including the new members joined after you added the listener.

    Messages are ordered, meaning that listeners(subscribers) will process the messages in the order they are actually
    published.

    """
    def add_listener(self, on_message=None):
        """
        Subscribes to this topic. When someone publishes a message on this topic, on_message() function is called if
        provided.

        :param on_message: (Function), function to be called when a message is published.
        :return: (str), a registration id which is used as a key to remove the listener.
        """
        request = topic_add_message_listener_codec.encode_request(self.name, self._is_smart)

        def handle(item, publish_time, uuid):
            member = self._client.cluster.get_member_by_uuid(uuid)
            item_event = TopicMessage(self.name, item, publish_time, member, self._to_object)
            on_message(item_event)

        return self._register_listener(
            request, lambda r: topic_add_message_listener_codec.decode_response(r)['response'],
            lambda reg_id: topic_remove_message_listener_codec.encode_request(self.name, reg_id),
            lambda m: topic_add_message_listener_codec.handle(m, handle))

    def publish(self, message):
        """
        Publishes the message to all subscribers of this topic

        :param message: (object), the message to be published.
        """
        message_data = self._to_data(message)
        self._encode_invoke(topic_publish_codec, message=message_data)

    def remove_listener(self, registration_id):
        """
        Stops receiving messages for the given message listener. If the given listener already removed, this method does
        nothing.

        :param registration_id: (str), registration id of the listener to be removed.
        :return: (bool), ``true`` if the listener is removed, ``false`` otherwise.
        """
        return self._deregister_listener(registration_id)



from hazelcast.protocol.codec import \
    set_add_all_codec, \
    set_add_codec, \
    set_add_listener_codec, \
    set_clear_codec, \
    set_compare_and_remove_all_codec, \
    set_compare_and_retain_all_codec, \
    set_contains_all_codec, \
    set_contains_codec, \
    set_get_all_codec, \
    set_is_empty_codec, \
    set_remove_codec, \
    set_remove_listener_codec, \
    set_size_codec

from hazelcast.proxy.base import PartitionSpecificProxy, ItemEvent, ItemEventType
from hazelcast.util import check_not_none


class Set(PartitionSpecificProxy):
    """
    Concurrent, distributed implementation of ``Set``.
    """
    def add(self, item):
        """
        Adds the specified item if it is not exists in this set.

        :param item: (object), the specified item to be added.
        :return: (bool), ``true`` if this set is changed after call, ``false`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)
        return self._encode_invoke(set_add_codec, value=element_data)

    def add_all(self, items):
        """
        Adds the elements in the specified collection if they're not exist in this set.

        :param items: (Collection), collection which includes the items to be added.
        :return: (bool), ``true`` if this set is changed after call, ``false`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(set_add_all_codec, value_list=data_items)

    def add_listener(self, include_value=False, item_added_func=None, item_removed_func=None):
        """
        Adds an item listener for this container. Listener will be notified for all container add/remove events.

        :param include_value: (bool), whether received events include the updated item or not (optional).
        :param item_added_func: Function to be called when an item is added to this set (optional).
        :param item_removed_func: Function to be called when an item is deleted from this set (optional).
        :return: (str), a registration id which is used as a key to remove the listener.
        """
        request = set_add_listener_codec.encode_request(self.name, include_value, self._is_smart)

        def handle_event_item(item, uuid, event_type):
            item = item if include_value else None
            member = self._client.cluster.get_member_by_uuid(uuid)

            item_event = ItemEvent(self.name, item, event_type, member, self._to_object)
            if event_type == ItemEventType.added:
                if item_added_func:
                    item_added_func(item_event)
            else:
                if item_removed_func:
                    item_removed_func(item_event)

        return self._register_listener(request, lambda r: set_add_listener_codec.decode_response(r)['response'],
                                       lambda reg_id: set_remove_listener_codec.encode_request(self.name, reg_id),
                                       lambda m: set_add_listener_codec.handle(m, handle_event_item))

    def clear(self):
        """
        Clears the set. Set will be empty with this call.
        """
        return self._encode_invoke(set_clear_codec)

    def contains(self, item):
        """
        Determines whether this set contains the specified item or not.

        :param item: (object), the specified item to be searched.
        :return: (bool), ``true`` if the specified item exists in this set, ``false`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)
        return self._encode_invoke(set_contains_codec, value=item_data)

    def contains_all(self, items):
        """
        Determines whether this set contains all of the items in the specified collection or not.

        :param items: (Collection), the specified collection which includes the items to be searched.
        :return: (bool), ``true`` if all of the items in the specified collection exist in this set, ``false`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(set_contains_all_codec, items=data_items)

    def get_all(self):
        """
        Returns all of the items in the set.

        :return: (Sequence), list of the items in this set.
        """
        return self._encode_invoke(set_get_all_codec)

    def is_empty(self):
        """
        Determines whether this set is empty or not.

        :return: (bool), ``true`` if this set is empty, ``false`` otherwise.
        """
        return self._encode_invoke(set_is_empty_codec)

    def remove(self, item):
        """
        Removes the specified element from the set if it exists.

        :param item: (object), the specified element to be removed.
        :return: (bool), ``true`` if the specified element exists in this set.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)
        return self._encode_invoke(set_remove_codec, value=item_data)

    def remove_all(self, items):
        """
        Removes all of the elements of the specified collection from this set.

        :param items: (Collection), the specified collection.
        :return: (bool), ``true`` if the call changed this set, ``false`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(set_compare_and_remove_all_codec, values=data_items)

    def remove_listener(self, registration_id):
        """
        Removes the specified item listener. Returns silently if the specified listener was not added before.

        :param registration_id: (str), id of the listener to be deleted.
        :return: (bool), ``true`` if the item listener is removed, ``false`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def retain_all(self, items):
        """
        Removes the items which are not contained in the specified collection. In other words, only the items that are
        contained in the specified collection will be retained.

        :param items: (Collection), collection which includes the elements to be retained in this set.
        :return: (bool), ``true`` if this set changed as a result of the call.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(set_compare_and_retain_all_codec, values=data_items)

    def size(self):
        """
        Returns the number of items in this set.

        :return: (int), number of items in this set.
        """
        return self._encode_invoke(set_size_codec)


from setuptools import setup, find_packages
from codecs import open
from os import path
from hazelcast import __version__
from hazelcast.version import write_git_info, get_commit_id, get_commit_date

here = path.abspath(path.dirname(__file__))

commit_id = get_commit_id()
commit_date = get_commit_date()
write_git_info(here, commit_id, commit_date)

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


stats_requirements = [
    'psutil',
]

extras = {
    'stats': stats_requirements,
}

setup(
        name='hazelcast-python-client',
        version=__version__,
        description='Hazelcast Python Client',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/hazelcast/hazelcast-python-client',

        author='Hazelcast Inc. Developers',
        author_email='hazelcast@googlegroups.com',

        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: Implementation :: CPython',
            # 'Programming Language :: Python :: Implementation :: PyPy',
            'Topic :: Software Development :: Libraries :: Python Modules'

        ],
        license='Apache 2.0',
        keywords='hazelcast,hazelcast client,In-Memory Data Grid,Distributed Computing',
        packages=find_packages(exclude=['benchmarks', 'examples', 'examples.*', 'docs', 'tests', 'tests.*']),
        package_dir={'hazelcast': 'hazelcast'},
        package_data={'hazelcast': ["git_info.json"]},
        install_requires=[],
        extras_require=extras,
        tests_require=['thrift', 'nose', 'coverage', 'psutil'],
)



import logging
import threading
from uuid import uuid4

from hazelcast.exception import OperationTimeoutError, HazelcastError
from hazelcast.util import current_time_in_millis, check_not_none
from time import sleep


class ListenerRegistration(object):
    def __init__(self, registration_request, decode_register_response, encode_deregister_request, handler):
        self.registration_request = registration_request
        self.decode_register_response = decode_register_response
        self.encode_deregister_request = encode_deregister_request
        self.handler = handler
        self.connection_registrations = {}  # Dict of Connection, EventRegistration


class EventRegistration(object):
    def __init__(self, server_registration_id, correlation_id):
        self.server_registration_id = server_registration_id
        self.correlation_id = correlation_id


class ListenerService(object):
    logger = logging.getLogger("HazelcastClient.ListenerService")

    def __init__(self, client):
        self._client = client
        self._invocation_service = client.invoker
        self.is_smart = client.config.network_config.smart_routing
        self._logger_extras = {"client_name": client.name, "group_name": client.config.group_config.name}
        self._active_registrations = {}  # Dict of user_registration_id, ListenerRegistration
        self._registration_lock = threading.RLock()
        self._event_handlers = {}

    def try_sync_connect_to_all_members(self):
        cluster_service = self._client.cluster
        start_millis = current_time_in_millis()
        while True:
            last_failed_member = None
            last_exception = None
            for member in cluster_service.members:
                try:
                    self._client.connection_manager.get_or_connect(member.address).result()
                except Exception as e:
                    last_failed_member = member
                    last_exception = e
            if last_exception is None:
                break
            self.time_out_or_sleep_before_next_try(start_millis, last_failed_member, last_exception)
            if not self._client.lifecycle.is_live():
                break

    def time_out_or_sleep_before_next_try(self, start_millis, last_failed_member, last_exception):
        now_in_millis = current_time_in_millis()
        elapsed_millis = now_in_millis - start_millis
        invocation_time_out_millis = self._invocation_service.invocation_timeout * 1000
        timed_out = elapsed_millis > invocation_time_out_millis
        if timed_out:
            raise OperationTimeoutError\
                ("Registering listeners is timed out. Last failed member: %s, Current time: %s, Start time: %s, "
                 "Client invocation timeout: %s, Elapsed time: %s ms, Cause: %s", last_failed_member, now_in_millis,
                 start_millis, invocation_time_out_millis, elapsed_millis, last_exception.args[0])
        else:
            sleep(self._invocation_service.invocation_retry_pause)  # sleep before next try

    def register_listener(self, registration_request, decode_register_response, encode_deregister_request, handler):
        if self.is_smart:
            self.try_sync_connect_to_all_members()

        with self._registration_lock:
            user_registration_id = str(uuid4())
            listener_registration = ListenerRegistration(registration_request, decode_register_response,
                                                         encode_deregister_request, handler)
            self._active_registrations[user_registration_id] = listener_registration

            active_connections = self._client.connection_manager.connections
            for connection in active_connections.values():
                try:
                    self.register_listener_on_connection_async(user_registration_id, listener_registration, connection)\
                        .result()
                except:
                    if connection.live():
                        self.deregister_listener(user_registration_id)
                        raise HazelcastError("Listener cannot be added ")
            return user_registration_id

    def register_listener_on_connection_async(self, user_registration_id, listener_registration, connection):
        registration_map = listener_registration.connection_registrations

        if connection in registration_map:
            return

        registration_request = listener_registration.registration_request.clone()
        future = self._invocation_service.invoke_on_connection(registration_request, connection,
                                                               event_handler=listener_registration.handler)

        def callback(f):
            try:
                response = f.result()
                server_registration_id = listener_registration.decode_register_response(response)
                correlation_id = registration_request.get_correlation_id()
                registration = EventRegistration(server_registration_id, correlation_id)
                registration_map[connection] = registration
            except Exception as e:
                if connection.live():
                    self.logger.warning("Listener %s can not be added to a new connection: %s, reason: %s",
                                        user_registration_id, connection, e.args[0], extra=self._logger_extras)
                raise e

        return future.continue_with(callback)

    def deregister_listener(self, user_registration_id):
        check_not_none(user_registration_id, "None userRegistrationId is not allowed!")

        with self._registration_lock:
            listener_registration = self._active_registrations.get(user_registration_id)
            if listener_registration is None:
                return False
            successful = True
            for connection, event_registration in list(listener_registration.connection_registrations.items()):
                try:
                    server_registration_id = event_registration.server_registration_id
                    deregister_request = listener_registration.encode_deregister_request(server_registration_id)
                    self._invocation_service.invoke_on_connection(deregister_request, connection).result()
                    self.remove_event_handler(event_registration.correlation_id)
                    listener_registration.connection_registrations.pop(connection)
                except:
                    if connection.live():
                        successful = False
                        self.logger.warning("Deregistration for listener with ID %s has failed to address %s ",
                                            user_registration_id, "address", exc_info=True, extra=self._logger_extras)
            if successful:
                self._active_registrations.pop(user_registration_id)
            return successful

    def connection_added(self, connection):
        with self._registration_lock:
            for user_reg_id, listener_registration in self._active_registrations.items():
                self.register_listener_on_connection_async(user_reg_id, listener_registration, connection)

    def connection_removed(self, connection, _):
        with self._registration_lock:
            for listener_registration in self._active_registrations.values():
                event_registration = listener_registration.connection_registrations.pop(connection, None)
                if event_registration is not None:
                    self.remove_event_handler(event_registration.correlation_id)

    def start(self):
        self._client.connection_manager.add_listener(self.connection_added, self.connection_removed)

    def handle_client_message(self, message):
        correlation_id = message.get_correlation_id()
        if correlation_id not in self._event_handlers:
            self.logger.warning("Got event message with unknown correlation id: %s", message, extra=self._logger_extras)
        else:
            event_handler = self._event_handlers.get(correlation_id)
            event_handler(message)

    def add_event_handler(self, correlation_id, event_handler):
        self._event_handlers[correlation_id] = event_handler

    def remove_event_handler(self, correlation_id):
        self._event_handlers.pop(correlation_id, None)



MAP_PUT = 0x0101
MAP_GET = 0x0102
MAP_REMOVE = 0x0103
MAP_REPLACE = 0x0104
MAP_REPLACEIFSAME = 0x0105
MAP_CONTAINSKEY = 0x0109
MAP_CONTAINSVALUE = 0x010a
MAP_REMOVEIFSAME = 0x010b
MAP_DELETE = 0x010c
MAP_FLUSH = 0x010d
MAP_TRYREMOVE = 0x010e
MAP_TRYPUT = 0x010f
MAP_PUTTRANSIENT = 0x0110
MAP_PUTIFABSENT = 0x0111
MAP_SET = 0x0112
MAP_LOCK = 0x0113
MAP_TRYLOCK = 0x0114
MAP_ISLOCKED = 0x0115
MAP_UNLOCK = 0x0116
MAP_ADDINTERCEPTOR = 0x0117
MAP_REMOVEINTERCEPTOR = 0x0118
MAP_ADDENTRYLISTENERTOKEYWITHPREDICATE = 0x0119
MAP_ADDENTRYLISTENERWITHPREDICATE = 0x011a
MAP_ADDENTRYLISTENERTOKEY = 0x011b
MAP_ADDENTRYLISTENER = 0x011c
MAP_ADDNEARCACHEENTRYLISTENER = 0x011d
MAP_REMOVEENTRYLISTENER = 0x011e
MAP_ADDPARTITIONLOSTLISTENER = 0x011f
MAP_REMOVEPARTITIONLOSTLISTENER = 0x0120
MAP_GETENTRYVIEW = 0x0121
MAP_EVICT = 0x0122
MAP_EVICTALL = 0x0123
MAP_LOADALL = 0x0124
MAP_LOADGIVENKEYS = 0x0125
MAP_KEYSET = 0x0126
MAP_GETALL = 0x0127
MAP_VALUES = 0x0128
MAP_ENTRYSET = 0x0129
MAP_KEYSETWITHPREDICATE = 0x012a
MAP_VALUESWITHPREDICATE = 0x012b
MAP_ENTRIESWITHPREDICATE = 0x012c
MAP_ADDINDEX = 0x012d
MAP_SIZE = 0x012e
MAP_ISEMPTY = 0x012f
MAP_PUTALL = 0x0130
MAP_CLEAR = 0x0131
MAP_EXECUTEONKEY = 0x0132
MAP_SUBMITTOKEY = 0x0133
MAP_EXECUTEONALLKEYS = 0x0134
MAP_EXECUTEWITHPREDICATE = 0x0135
MAP_EXECUTEONKEYS = 0x0136
MAP_FORCEUNLOCK = 0x0137
MAP_KEYSETWITHPAGINGPREDICATE = 0x0138
MAP_VALUESWITHPAGINGPREDICATE = 0x0139
MAP_ENTRIESWITHPAGINGPREDICATE = 0x013a
MAP_CLEARNEARCACHE = 0x013b



from hazelcast.protocol.codec import transactional_list_add_codec, transactional_list_remove_codec, \
    transactional_list_size_codec
from hazelcast.proxy.base import TransactionalProxy
from hazelcast.util import check_not_none


class TransactionalList(TransactionalProxy):
    """
    Transactional implementation of :class:`~hazelcast.proxy.list.List`.
    """
    def add(self, item):
        """
        Transactional implementation of :func:`List.add(item) <hazelcast.proxy.list.List.add>`

        :param item: (object), the new item to be added.
        :return: (bool), ``true`` if the item is added successfully, ``false`` otherwise.
        """
        check_not_none(item, "item can't be none")
        return self._encode_invoke(transactional_list_add_codec, item=self._to_data(item))

    def remove(self, item):
        """
        Transactional implementation of :func:`List.remove(item) <hazelcast.proxy.list.List.remove>`

        :param item: (object), the specified item to be removed.
        :return: (bool), ``true`` if the item is removed successfully, ``false`` otherwise.
        """
        check_not_none(item, "item can't be none")
        return self._encode_invoke(transactional_list_remove_codec, item=self._to_data(item))

    def size(self):
        """
        Transactional implementation of :func:`List.size() <hazelcast.proxy.list.List.size>`

        :return: (int), the size of the list.
        """
        return self._encode_invoke(transactional_list_size_codec)



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.codec.replicated_map_message_type import *
from hazelcast.protocol.event_response_const import *

REQUEST_TYPE = REPLICATEDMAP_ADDNEARCACHEENTRYLISTENER
RESPONSE_TYPE = 104
RETRYABLE = False


def calculate_size(name, include_value, local_only):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(name)
    data_size += BOOLEAN_SIZE_IN_BYTES
    data_size += BOOLEAN_SIZE_IN_BYTES
    return data_size


def encode_request(name, include_value, local_only):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(name, include_value, local_only))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(name)
    client_message.append_bool(include_value)
    client_message.append_bool(local_only)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(response=None)
    parameters['response'] = client_message.read_str()
    return parameters


def handle(client_message, handle_event_entry=None, to_object=None):
    """ Event handler """
    message_type = client_message.get_message_type()
    if message_type == EVENT_ENTRY and handle_event_entry is not None:
        key = None
        if not client_message.read_bool():
            key = client_message.read_data()
        value = None
        if not client_message.read_bool():
            value = client_message.read_data()
        old_value = None
        if not client_message.read_bool():
            old_value = client_message.read_data()
        merging_value = None
        if not client_message.read_bool():
            merging_value = client_message.read_data()
        event_type = client_message.read_int()
        uuid = client_message.read_str()
        number_of_affected_entries = client_message.read_int()
        handle_event_entry(key=key, value=value, old_value=old_value, merging_value=merging_value, event_type=event_type, uuid=uuid, number_of_affected_entries=number_of_affected_entries)




import logging
import uuid

from hazelcast.util import create_git_info

LIFECYCLE_STATE_STARTING = "STARTING"
LIFECYCLE_STATE_CONNECTED = "CONNECTED"
LIFECYCLE_STATE_DISCONNECTED = "DISCONNECTED"
LIFECYCLE_STATE_SHUTTING_DOWN = "SHUTTING_DOWN"
LIFECYCLE_STATE_SHUTDOWN = "SHUTDOWN"


class LifecycleService(object):
    """
    LifecycleService allows you to shutdown, terminate, and listen to LifecycleEvent's on HazelcastInstances.
    """
    logger = logging.getLogger("HazelcastClient.LifecycleService")
    state = None

    def __init__(self, config, logger_extras=None):
        self._listeners = {}
        self._logger_extras = logger_extras

        for listener in config.lifecycle_listeners:
            self.add_listener(listener)

        self._git_info = create_git_info()
        self.is_live = True
        self.fire_lifecycle_event(LIFECYCLE_STATE_STARTING)

    def add_listener(self, on_lifecycle_change):
        """
        Add a listener object to listen for lifecycle events.

        :param on_lifecycle_change: (Function), function to be called when LifeCycle state is changed.
        :return: (str), id of the listener.
        """
        id = str(uuid.uuid4())
        self._listeners[id] = on_lifecycle_change
        return id

    def remove_listener(self, registration_id):
        """
        Removes a lifecycle listener.

        :param registration_id: (str), the id of the listener to be removed.
        :return: (bool), ``true`` if the listener is removed successfully, ``false`` otherwise.
        """
        try:
            self._listeners.pop(registration_id)
            return True
        except KeyError:
            return False

    def fire_lifecycle_event(self, new_state):
        """
        Called when instance's state changes.

        :param new_state: (Lifecycle State), the new state of the instance.
        """
        if new_state == LIFECYCLE_STATE_SHUTTING_DOWN:
            self.is_live = False

        self.state = new_state
        self.logger.info(self._git_info + "HazelcastClient is %s", new_state, extra=self._logger_extras)
        for listener in list(self._listeners.values()):
            try:
                listener(new_state)
            except:
                self.logger.exception("Exception in lifecycle listener", extra=self._logger_extras)



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.util import ImmutableLazyDataList
from hazelcast.protocol.codec.map_message_type import *
from hazelcast.six.moves import range

REQUEST_TYPE = MAP_EXECUTEONALLKEYS
RESPONSE_TYPE = 117
RETRYABLE = False


def calculate_size(name, entry_processor):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(name)
    data_size += calculate_size_data(entry_processor)
    return data_size


def encode_request(name, entry_processor):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(name, entry_processor))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(name)
    client_message.append_data(entry_processor)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(response=None)
    response_size = client_message.read_int()
    response = []
    for _ in range(0, response_size):
        response_item = (client_message.read_data(), client_message.read_data())
        response.append(response_item)
    parameters['response'] = ImmutableLazyDataList(response, to_object)
    return parameters



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.util import ImmutableLazyDataList
from hazelcast.protocol.codec.ringbuffer_message_type import *
from hazelcast.six.moves import range

REQUEST_TYPE = RINGBUFFER_READMANY
RESPONSE_TYPE = 115
RETRYABLE = True


def calculate_size(name, start_sequence, min_count, max_count, filter):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(name)
    data_size += LONG_SIZE_IN_BYTES
    data_size += INT_SIZE_IN_BYTES
    data_size += INT_SIZE_IN_BYTES
    data_size += BOOLEAN_SIZE_IN_BYTES
    if filter is not None:
        data_size += calculate_size_data(filter)
    return data_size


def encode_request(name, start_sequence, min_count, max_count, filter):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(name, start_sequence, min_count, max_count, filter))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(name)
    client_message.append_long(start_sequence)
    client_message.append_int(min_count)
    client_message.append_int(max_count)
    client_message.append_bool(filter is None)
    if filter is not None:
        client_message.append_data(filter)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(read_count=None, items=None)
    parameters['read_count'] = client_message.read_int()
    items_size = client_message.read_int()
    items = []
    for _ in range(0, items_size):
        items_item = client_message.read_data()
        items.append(items_item)
    parameters['items'] = ImmutableLazyDataList(items, to_object)
    return parameters



from hazelcast.exception import HazelcastSerializationError
from hazelcast.serialization.base import BaseSerializationService
from hazelcast.serialization.portable.classdef import FieldType
from hazelcast.serialization.portable.context import PortableContext
from hazelcast.serialization.portable.serializer import PortableSerializer
from hazelcast.serialization.serializer import *
from hazelcast import six
from hazelcast.config import ClientProperties

DEFAULT_OUT_BUFFER_SIZE = 4 * 1024


def default_partition_strategy(key):
    if hasattr(key, "get_partition_key"):
        return key.get_partition_key()
    return None


class SerializationServiceV1(BaseSerializationService):

    def __init__(self, serialization_config, properties=ClientProperties({}), version=1, global_partition_strategy=default_partition_strategy,
                 output_buffer_size=DEFAULT_OUT_BUFFER_SIZE):
        super(SerializationServiceV1, self).__init__(version, global_partition_strategy, output_buffer_size,
                                                     serialization_config.is_big_endian,
                                                     serialization_config.default_integer_type)
        self._portable_context = PortableContext(self, serialization_config.portable_version)
        self.register_class_definitions(serialization_config.class_definitions, serialization_config.check_class_def_errors)
        self._registry._portable_serializer = PortableSerializer(self._portable_context, serialization_config.portable_factories)

        # merge configured factories with built in ones
        factories = {}
        factories.update(serialization_config.data_serializable_factories)
        self._registry._data_serializer = IdentifiedDataSerializer(factories)
        self._register_constant_serializers()

        # Register Custom Serializers
        for _type, custom_serializer in six.iteritems(serialization_config.custom_serializers):
            self._registry.safe_register_serializer(custom_serializer(), _type)

        # Register Global Serializer
        global_serializer = serialization_config.global_serializer
        if global_serializer:
            self._registry._global_serializer = global_serializer()

        self.properties = properties

    def _register_constant_serializers(self):
        self._registry.register_constant_serializer(self._registry._null_serializer, type(None))
        self._registry.register_constant_serializer(self._registry._data_serializer)
        self._registry.register_constant_serializer(self._registry._portable_serializer)
        self._registry.register_constant_serializer(ByteSerializer())
        self._registry.register_constant_serializer(BooleanSerializer(), bool)
        self._registry.register_constant_serializer(CharSerializer())
        self._registry.register_constant_serializer(ShortSerializer())
        self._registry.register_constant_serializer(IntegerSerializer())
        self._registry.register_constant_serializer(LongSerializer())
        self._registry.register_constant_serializer(FloatSerializer())
        self._registry.register_constant_serializer(DoubleSerializer(), float)
        self._registry.register_constant_serializer(StringSerializer())
        # Arrays of primitives and String
        self._registry.register_constant_serializer(ByteArraySerializer())
        self._registry.register_constant_serializer(BooleanArraySerializer())
        self._registry.register_constant_serializer(CharArraySerializer())
        self._registry.register_constant_serializer(ShortArraySerializer())
        self._registry.register_constant_serializer(IntegerArraySerializer())
        self._registry.register_constant_serializer(LongArraySerializer())
        self._registry.register_constant_serializer(FloatArraySerializer())
        self._registry.register_constant_serializer(DoubleArraySerializer())
        self._registry.register_constant_serializer(StringArraySerializer())
        # EXTENSIONS
        self._registry.register_constant_serializer(DateTimeSerializer(), datetime)
        self._registry.register_constant_serializer(BigIntegerSerializer())
        self._registry.register_constant_serializer(BigDecimalSerializer())
        self._registry.register_constant_serializer(JavaClassSerializer())
        self._registry.register_constant_serializer(JavaEnumSerializer())
        self._registry.register_constant_serializer(ArrayListSerializer(), list)
        self._registry.register_constant_serializer(LinkedListSerializer())
        self._registry.register_constant_serializer(HazelcastJsonValueSerializer(), HazelcastJsonValue)

        self._registry.safe_register_serializer(self._registry._python_serializer)

    def register_class_definitions(self, class_definitions, check_error):
        class_defs = dict()
        for cd in class_definitions:
            if cd in class_defs:
                raise HazelcastSerializationError("Duplicate registration found for class-id:{}".format(cd.class_id))
            class_defs[cd.class_id] = cd
        for cd in class_definitions:
            self.register_class_definition(cd, class_defs, check_error)

    def register_class_definition(self, cd, class_defs, check_error):
        field_names = cd.get_field_names()
        for field_name in field_names:
            fd = cd.get_field(field_name)
            if fd.field_type == FieldType.PORTABLE or fd.field_type == FieldType.PORTABLE_ARRAY:
                nested_cd = class_defs.get(fd.class_id, None)
                if nested_cd is not None:
                    self.register_class_definition(nested_cd, class_defs, check_error)
                    self._portable_context.register_class_definition(nested_cd)
                elif check_error:
                    raise HazelcastSerializationError(
                            "Could not find registered ClassDefinition for class-id:{}".format(fd.class_id))
        self._portable_context.register_class_definition(cd)




ATOMICLONG_APPLY = 0x0a01
ATOMICLONG_ALTER = 0x0a02
ATOMICLONG_ALTERANDGET = 0x0a03
ATOMICLONG_GETANDALTER = 0x0a04
ATOMICLONG_ADDANDGET = 0x0a05
ATOMICLONG_COMPAREANDSET = 0x0a06
ATOMICLONG_DECREMENTANDGET = 0x0a07
ATOMICLONG_GET = 0x0a08
ATOMICLONG_GETANDADD = 0x0a09
ATOMICLONG_GETANDSET = 0x0a0a
ATOMICLONG_INCREMENTANDGET = 0x0a0b
ATOMICLONG_GETANDINCREMENT = 0x0a0c
ATOMICLONG_SET = 0x0a0d



from random import randint
from hazelcast.protocol.codec import replicated_map_clear_codec, replicated_map_add_entry_listener_codec, \
    replicated_map_add_entry_listener_to_key_codec, replicated_map_add_entry_listener_to_key_with_predicate_codec, \
    replicated_map_add_entry_listener_with_predicate_codec, replicated_map_contains_key_codec, \
    replicated_map_contains_value_codec, replicated_map_entry_set_codec, replicated_map_get_codec, \
    replicated_map_is_empty_codec, replicated_map_key_set_codec, replicated_map_put_all_codec, replicated_map_put_codec, \
    replicated_map_remove_codec, replicated_map_remove_entry_listener_codec, replicated_map_size_codec, \
    replicated_map_values_codec
from hazelcast.proxy.base import Proxy, default_response_handler, EntryEvent, EntryEventType
from hazelcast.util import to_millis, check_not_none
from hazelcast import six


class ReplicatedMap(Proxy):
    """
    A ReplicatedMap is a map-like data structure with weak consistency and values locally stored on every node of the
    cluster.

    Whenever a value is written asynchronously, the new value will be internally distributed to all existing cluster
    members, and eventually every node will have the new value.

    When a new node joins the cluster, the new node initially will request existing values from older nodes and
    replicate them locally.
    """
    _partition_id = None

    def add_entry_listener(self, key=None, predicate=None, added_func=None, removed_func=None, updated_func=None,
                           evicted_func=None, clear_all_func=None):
        """
        Adds a continuous entry listener for this map. Listener will get notified for map events filtered with given
        parameters.

        :param key: (object), key for filtering the events (optional).
        :param predicate: (Predicate), predicate for filtering the events (optional).
        :param added_func: Function to be called when an entry is added to map (optional).
        :param removed_func: Function to be called when an entry is removed from map (optional).
        :param updated_func: Function to be called when an entry is updated (optional).
        :param evicted_func: Function to be called when an entry is evicted from map (optional).
        :param clear_all_func: Function to be called when entries are cleared from map (optional).
        :return: (str), a registration id which is used as a key to remove the listener.

        .. seealso:: :class:`~hazelcast.serialization.predicate.Predicate` for more info about predicates.
        """
        if key and predicate:
            key_data = self._to_data(key)
            predicate_data = self._to_data(predicate)
            request = replicated_map_add_entry_listener_to_key_with_predicate_codec.encode_request(
                self.name, key_data, predicate_data, self._is_smart)
        elif key and not predicate:
            key_data = self._to_data(key)
            request = replicated_map_add_entry_listener_to_key_codec.encode_request(self.name, key_data, self._is_smart)
        elif not key and predicate:
            predicate = self._to_data(predicate)
            request = replicated_map_add_entry_listener_with_predicate_codec.encode_request(
                self.name, predicate, self._is_smart)
        else:
            request = replicated_map_add_entry_listener_codec.encode_request(self.name, self._is_smart)

        def handle_event_entry(**_kwargs):
            event = EntryEvent(self._to_object, **_kwargs)
            if event.event_type == EntryEventType.added and added_func:
                added_func(event)
            elif event.event_type == EntryEventType.removed and removed_func:
                removed_func(event)
            elif event.event_type == EntryEventType.updated and updated_func:
                updated_func(event)
            elif event.event_type == EntryEventType.evicted and evicted_func:
                evicted_func(event)
            elif event.event_type == EntryEventType.clear_all and clear_all_func:
                clear_all_func(event)

        return self._register_listener(
            request, lambda r: replicated_map_add_entry_listener_codec.decode_response(r)['response'],
            lambda reg_id: replicated_map_remove_entry_listener_codec.encode_request(self.name, reg_id),
            lambda m: replicated_map_add_entry_listener_codec.handle(m, handle_event_entry))

    def clear(self):
        """
        Wipes data out of the replicated map.
        """
        return self._encode_invoke(replicated_map_clear_codec)

    def contains_key(self, key):
        """
        Determines whether this map contains an entry with the key.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), the specified key.
        :return: (bool), ``true`` if this map contains an entry for the specified key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._encode_invoke_on_key(replicated_map_contains_key_codec, key_data, key=key_data)

    def contains_value(self, value):
        """
        Determines whether this map contains one or more keys for the specified value.

        :param value: (object), the specified value.
        :return: (bool), ``true`` if this map contains an entry for the specified value.
        """
        check_not_none(value, "value can't be None")
        return self._encode_invoke_on_target_partition(replicated_map_contains_value_codec, value=self._to_data(value))

    def entry_set(self):
        """
        Returns a List clone of the mappings contained in this map.

        **Warning:
        The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.**

        :return: (Sequence), the list of key-value tuples in the map.
        """
        return self._encode_invoke_on_target_partition(replicated_map_entry_set_codec)

    def get(self, key):
        """
        Returns the value for the specified key, or None if this map does not contain this key.

        **Warning:
        This method uses hashCode and equals of the binary form of the key, not the actual implementations of hashCode
        and equals defined in the key's class.**

        :param key: (object), the specified key.
        :return: (Sequence), the list of the values associated with the specified key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._encode_invoke_on_key(replicated_map_get_codec, key_data, key=key_data)

    def is_empty(self):
        """
        Returns ``true`` if this map contains no key-value mappings.

        :return: (bool), ``true`` if this map contains no key-value mappings.
        """
        return self._encode_invoke_on_target_partition(replicated_map_is_empty_codec)

    def key_set(self):
        """
        Returns the list of keys in the ReplicatedMap.

        **Warning:
        The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.**

        :return: (Sequence), a list of the clone of the keys.
        """
        return self._encode_invoke_on_target_partition(replicated_map_key_set_codec)

    def put(self, key, value, ttl=0):
        """
        Associates the specified value with the specified key in this map. If the map previously contained a mapping for
        the key, the old value is replaced by the specified value. If ttl is provided, entry will expire and get evicted
        after the ttl.

        :param key: (object), the specified key.
        :param value: (object), the value to associate with the key.
        :param ttl: (int), maximum time in seconds for this entry to stay, if not provided, the value configured on
            server side configuration will be used(optional).
        :return: (object), previous value associated with key or None if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        check_not_none(key, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._encode_invoke_on_key(replicated_map_put_codec, key_data, key=key_data, value=value_data,
                                          ttl=to_millis(ttl))

    def put_all(self, map):
        """
        Copies all of the mappings from the specified map to this map. No atomicity guarantees are
        given. In the case of a failure, some of the key-value tuples may get written, while others are not.

        :param map: (dict), map which includes mappings to be stored in this map.
        """
        entries = {}
        for key, value in six.iteritems(map):
            check_not_none(key, "key can't be None")
            check_not_none(value, "value can't be None")
            entries[self._to_data(key)] = self._to_data(value)
        self._encode_invoke(replicated_map_put_all_codec, entries=entries)

    def remove(self, key):
        """
        Removes the mapping for a key from this map if it is present. The map will not contain a mapping for the
        specified key once the call returns.

        **Warning: This method uses __hash__ and __eq__ methods of binary form of the key, not the actual implementations
        of __hash__ and __eq__ defined in key's class.**

        :param key: (object), key of the mapping to be deleted.
        :return: (object), the previous value associated with key, or None if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._encode_invoke_on_key(replicated_map_remove_codec, key_data, key=key_data)

    def remove_entry_listener(self, registration_id):
        """
        Removes the specified entry listener. Returns silently if there is no such listener added before.

        :param registration_id: (str), id of registered listener.
        :return: (bool), ``true`` if registration is removed, ``false`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def size(self):
        """
        Returns the number of entries in this multimap.

        :return: (int), number of entries in this multimap.
        """
        return self._encode_invoke_on_target_partition(replicated_map_size_codec)

    def values(self):
        """
        Returns the list of values in the map.

        **Warning:
        The returned list is NOT backed by the map, so changes to the map are NOT reflected in the list, and
        vice-versa.**

        :return: (Sequence), the list of values in the map.
        """
        return self._encode_invoke_on_target_partition(replicated_map_values_codec)

    def _get_partition_id(self):
        if not self._partition_id:
            self._partition_id = randint(0, self._client.partition_service.get_partition_count() - 1)
        return self._partition_id

    def _encode_invoke_on_target_partition(self, codec, response_handler=default_response_handler, **kwargs):
        return self._encode_invoke_on_partition(codec, self._get_partition_id(), response_handler, **kwargs)



from hazelcast.protocol.codec import \
    count_down_latch_await_codec, \
    count_down_latch_count_down_codec, \
    count_down_latch_get_count_codec, \
    count_down_latch_try_set_count_codec

from hazelcast.proxy.base import PartitionSpecificProxy
from hazelcast.util import check_not_negative, to_millis
from hazelcast import six


class CountDownLatch(PartitionSpecificProxy):
    """
    CountDownLatch is a backed-up, distributed, cluster-wide synchronization aid that allows one or more threads to wait until a
    set of operations being performed in other threads completes
    """
    if six.PY2:
        six.exec_("""def await(self, timeout):
            return self.await_latch(timeout)
        """)

    def await_latch(self, timeout):
        """
        Causes the current thread to wait until the latch has counted down to zero, or the specified waiting time
        elapses.

        If the current count is zero then this method returns immediately with the value ``true``.

        If the current count is greater than zero, then the current thread becomes disabled for thread scheduling
        purposes and lies dormant until one of following happens:

            * the count reaches zero due to invocations of the countDown() method,
            * this CountDownLatch instance is destroyed,
            * the countdown owner becomes disconnected,
            * some other thread interrupts the current thread, or
            * the specified waiting time elapses.

        If the count reaches zero, then the method returns with the value ``true``.

        :param timeout: (long), the maximum time in seconds to wait.
        :return: (bool), ``true`` if the count reached zero, ``false`` if the waiting time elapsed before the count reached zero.
        """
        return self._encode_invoke(count_down_latch_await_codec, timeout=to_millis(timeout))

    def count_down(self):
        """
        Decrements the count of the latch, releasing all waiting threads if the count reaches zero.

        If the current count is greater than zero, then it is decremented. If the new count is zero:
            * All waiting threads are re-enabled for thread scheduling purposes, and
            * Countdown owner is set to ``None``.

        If the current count equals zero, nothing happens.
        """
        return self._encode_invoke(count_down_latch_count_down_codec)

    def get_count(self):
        """
        Returns the current count

        :return: (int), the current count.
        """
        return self._encode_invoke(count_down_latch_get_count_codec)

    def try_set_count(self, count):
        """
        Sets the count to the given value if the current count is zero. If count is not zero, this method does nothing
        and returns ``false``.

        :param count: (int), the number of times count_down() must be invoked before threads can pass through await().
        :return: (bool), ``true`` if the new count was set, ``false`` if the current count is not zero.
        """
        check_not_negative(count, "count can't be negative")
        return self._encode_invoke(count_down_latch_try_set_count_codec, count=count)

